/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg;

import static org.fortiss.af3.exploration.alg.ExplorationAlgActivator.getDefault;
import static org.fortiss.af3.exploration.alg.service.ExplorationServiceManager.getService;

import org.eclipse.ui.IStartup;
import org.fortiss.af3.exploration.alg.backend.Opt4JDseBackend;
import org.fortiss.af3.exploration.alg.dse.evaluator.objective.MappingEvaluatorObjective;
import org.fortiss.af3.exploration.alg.dse.module.IEvaluatorModule;
import org.fortiss.af3.exploration.alg.feature.DesignDiversityExploration;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.alg.feature.PartitionArchitectureExploration;
import org.fortiss.af3.exploration.alg.feature.SafetyArchitectureExploration;
import org.fortiss.af3.exploration.alg.feature.TTScheduleExploration;
import org.fortiss.af3.exploration.alg.feature.TaskMappingExploration;
import org.fortiss.af3.exploration.alg.service.IExplorationConstraintTransformationService;
import org.fortiss.af3.exploration.alg.service.IExplorationEvaluatationService;
import org.fortiss.af3.exploration.alg.service.IExplorationFeatureService;
import org.fortiss.af3.exploration.alg.service.IExplorationRepairService;
import org.fortiss.af3.exploration.alg.service.IExplorationService;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService;
import org.fortiss.af3.exploration.alg.service.internal.ExplorationConstraintTransformationService;
import org.fortiss.af3.exploration.alg.service.internal.ExplorationEvaluationService;
import org.fortiss.af3.exploration.alg.service.internal.ExplorationFeatureService;
import org.fortiss.af3.exploration.alg.service.internal.ExplorationRepairService;
import org.fortiss.af3.exploration.alg.service.internal.ExplorationTransformationService;
import org.fortiss.af3.exploration.backend.IDSEBackendService;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.smt.model.SMTObjective;
import org.opt4j.core.DoubleValue;
import org.osgi.framework.BundleContext;

/**
 * Centralized class that provides the startup routines of all services provided by this plugin.
 * 
 * The order of execution is as follows:
 * - Initialize own services.
 * - Register with the exploration plugin. (Make the "Parent" aware of me)
 * - Start own services.
 * 
 * @author diewald
 */
public final class ExplorationAlg implements IStartup {

	/** {@inheritDoc} */
	@Override
	public void earlyStartup() {
		initialize();
		registerBackend();
		start();
	}

	/**
	 * Initializes the {@link IExplorationService}s contained in this plugin (first service
	 * activation stage).
	 */
	public static void initialize() {
		registerServices();

		getService(IExplorationFeatureService.class).initializeService();
		getService(IExplorationConstraintTransformationService.class).initializeService();
		getService(IExplorationEvaluatationService.class).initializeService();
		getService(IExplorationRepairService.class).initializeService();
		getService(IExplorationTransformationService.class).initializeService();

		// Initializes/Registers the bundled exploration modules.
		registerExplorationFeatures();

		// Add addtional Evaluators, not specific for a particular exploration feature.
		registerEvaluators();
	}

	/** Registers the provided services with the {@link BundleContext}. */
	private static void registerServices() {
		BundleContext context = getDefault().getBundle().getBundleContext();

		context.registerService(IExplorationFeatureService.class.getName(),
				ExplorationFeatureService.getInstance(), null);
		context.registerService(IExplorationConstraintTransformationService.class.getName(),
				ExplorationConstraintTransformationService.getInstance(), null);
		context.registerService(IExplorationEvaluatationService.class.getName(),
				ExplorationEvaluationService.getInstance(), null);
		context.registerService(IExplorationRepairService.class.getName(),
				ExplorationRepairService.getInstance(), null);
		context.registerService(IExplorationTransformationService.class.getName(),
				ExplorationTransformationService.getInstance(), null);
	}

	/** Registers the exploration modules bundled with this plugin. */
	private static void registerExplorationFeatures() {
		// Functional groups of Decoders.
		getService(IExplorationFeatureService.class)
				.registerExplorationFeature(new SafetyArchitectureExploration());
		getService(IExplorationFeatureService.class)
				.registerExplorationFeature(new PartitionArchitectureExploration());
		getService(IExplorationFeatureService.class)
				.registerExplorationFeature(new TaskMappingExploration());
		getService(IExplorationFeatureService.class)
				.registerExplorationFeature(new DesignDiversityExploration());
		getService(IExplorationFeatureService.class)
				.registerExplorationFeature(new TTScheduleExploration());
	}

	/**
	 * Registers {@link IEvaluatorModule}s that are not specific for some
	 * {@link IExplorationFeature}.
	 */
	@SuppressWarnings("unchecked")
	private static void registerEvaluators() {
		getService(IExplorationEvaluatationService.class)
				.registerEvaluator(new MappingEvaluatorObjective(
						(Class<? extends ExplorationTarget<DoubleValue>>)SMTObjective.class));
	}

	// TODO Perform registration in
	// org.fortiss.af3.rcp.application.advisors.ApplicationWorkbenchAdvisor (or rather an AF3DSE
	// module) once this plugin has been integrated into the AF3 product.
	/** Register the DSE backend with the exploration plugin. */
	private static void registerBackend() {
		IDSEBackendService.getInstance().registerDSEBackend(new Opt4JDseBackend());
	}

	/**
	 * Starts the {@link IExplorationService}s contained in this plugin (second service activation
	 * stage).
	 */
	public static void start() {
		// TODO: Check start order w.r.t. the transformation service.
		getService(IExplorationFeatureService.class).startService();
		getService(IExplorationConstraintTransformationService.class).startService();
		// The Transformation Service must be started before the Evaluation Service such that it can
		// check for the availability of required transformation modules.
		getService(IExplorationTransformationService.class).startService();
		getService(IExplorationEvaluatationService.class).startService();
		getService(IExplorationRepairService.class).startService();
	}
}
