/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickInstanceOf;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.fortiss.af3.exploration.alg.exception.ConstraintGenerationException;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationObjective;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.time.PeriodConstraint;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.af3.exploration.moea.model.predefined.PartitionMinObjective;
import org.fortiss.af3.exploration.moea.model.predefined.PredefinedFactory;
import org.fortiss.af3.exploration.util.CommonExplorationTagetModelElementFactory;

/**
 * Factory for creating implicit Exploration Targets based on the input system model.
 * 
 * @author diewald
 */
public class ImplicitExplorationTargetFactory {

	/** Private constructor to prevent instantiation from external (singleton). */
	private ImplicitExplorationTargetFactory() {
	}

	/**
	 * Creates "implicit" {@link ExplorationTarget}s from the {@link SystemModelAdapter} and adds
	 * them to the given {@link DseSpecification} model.
	 * "Implicit" {@link ExplorationTarget}s are defined by the system model itself and cannot be
	 * <i>directly</i> modified by the user.
	 * 
	 * @param dse
	 *            {@link DseSpecification} to which the "implicit" {@link ExplorationTarget}s
	 *            shall be added.
	 * @param systemModelAdapter
	 *            {@link SystemModelAdapter} from which the "implicit" {@link ExplorationTarget}s
	 *            are deduced.
	 * @return Collection of constraints generated from the model.
	 * @throws Exception
	 *             if any of the constraint generators fails.
	 */
	public static Collection<ExplorationTarget<?>> createImplicitExplorationTargets(
			DseSpecification dse, SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter)
			throws Exception {
		List<ExplorationTarget<?>> targets = new ArrayList<>();
		// constraints.addAll(createPeriodConstraints(systemModelAdapter));
		targets.add(createNoEmptyPartitionConstraint());
		targets.add(createSameSilInPartitionConstraint());
		targets.add(createPartitionMinObjective());

		// Reassign the ids of generated constraints. By default, they are set to 0. Each constraint
		// needs a "unique" id for its constraint type.
		int constrId = 0;
		for(ExplorationConstraint<?> curConstraint : pickInstanceOf(ExplorationConstraint.class,
				targets)) {
			curConstraint.setId(constrId);
			constrId++;
		}

		return targets;
	}

	/**
	 * Creates a single constraint that states a solution may not contain partitions without
	 * assigned tasks.
	 */
	private static ExplorationConstraint<?> createNoEmptyPartitionConstraint() {
		return PredefinedFactory.eINSTANCE.createNoEmptyPartitionConstraint();
	}

	/**
	 * Creates a single constraint that states a partition may only host tasks with the same SIL.
	 */
	private static ExplorationConstraint<?> createSameSilInPartitionConstraint() {
		return PredefinedFactory.eINSTANCE.createSameSilInPartitionConstraint();
	}

	/**
	 * Creates a {@link PeriodConstraint} for each {@link TaskGraph} present in the
	 * {@link SystemModelAdapter}.
	 * 
	 * @throws ConstraintGenerationException
	 *             if the generation of any {@link PeriodConstraint} fails.
	 */
	private static Collection<ExplorationConstraint<?>>
			createPeriodConstraints(SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter)
					throws ConstraintGenerationException {
		Collection<ExplorationConstraint<?>> constraints = new ArrayList<>();
		for(TaskGraph tg : systemModelAdapter.getTaskGraphs()) {
			for(ITaskAdapter<?> component : tg.getTasks()) {
				if(!(component.getObject() instanceof Component)) {
					throw new ConstraintGenerationException(PeriodConstraint.class,
							"Invalid model element type for this constraint: " +
									component.getObject().getClass().getSimpleName() +
									". Expected a Component.");
				}
				constraints.add(
						createPeriodConstraint((Component)component.getObject(), tg.getPeriod()));
			}
		}
		return constraints;
	}

	/**
	 * Factory method to create {@link PeriodConstraint}s.
	 * 
	 * @throws ConstraintGenerationException
	 *             if the period value is invalid ({@code null}, or <= 0).
	 */
	private static PeriodConstraint createPeriodConstraint(Component component, Double period)
			throws ConstraintGenerationException {
		if(component == null) {
			throw new ConstraintGenerationException(PeriodConstraint.class,
					"Cannot generate a PeriodConstraint for a 'null' Component.");
		}
		if(period == null || period <= 0.0) {
			throw new ConstraintGenerationException(PeriodConstraint.class,
					"Invalid period specified for the Component " + component.getName() + ".");
		}

		PeriodConstraint periodConstraint =
				CommonExplorationTagetModelElementFactory.createPeriodConstraint();
		periodConstraint.setName("Period Constraint(" + periodConstraint.getId() + ")");
		periodConstraint.setComponent(component);
		periodConstraint.setPeriod(period);
		return periodConstraint;
	}

	/**
	 * Creates an {@link ExplorationObjective} that aims to minimize the number of instantiated
	 * partitions.
	 */
	public static PartitionMinObjective createPartitionMinObjective() {
		PartitionMinObjective partMinObjective =
				PredefinedFactory.eINSTANCE.createPartitionMinObjective();
		partMinObjective.setName("Partition Count Minimization");
		return partMinObjective;
	}
}
