/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.IProblemDimension;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

/**
 * Base class to track status information of "individual evolution" during optimization.
 * 
 * @author diewald
 */
// TODO: transform this class into a status genotype the track the status information.
// NOTE: This Class is not used at the moment. It ONLY serves as a reference for the genotype
// implementation.
@SuppressWarnings("javadoc")
public class SolutionCandidateBase {

	private Collection<ExplorationConstraint<?>> constraints;

	private Map<IProblemDimension, Integer> constraintViolationsPerProblemDimension;

	private Map<IProblemDimension, Double> constraintViolationDegreePerProblemDimension;

	private Map<ExplorationConstraint<?>, Double> constraintViolationValue;

	private Map<ExplorationConstraint<?>, Double> constraintViolationValueTracker;

	private Map<ExplorationConstraint<?>, Integer> constraintViolationTracker;

	// TODO: replace this local eps-tolerance with a settable one
	private double violationTolerance = 1e-9;

	/** Constructor */
	public SolutionCandidateBase() {
		constraintViolationsPerProblemDimension = new HashMap<>();
		constraintViolationDegreePerProblemDimension = new HashMap<>();
		constraintViolationValue = new HashMap<>();
		constraintViolationValueTracker = new HashMap<>();
		constraintViolationTracker = new HashMap<>();
	}

	/**
	 * Adds the problem dimensions and constraints for a given {link GoalSpecification}. These are
	 * required to allow tracking and evaluating constraint violation of the possible solutions of
	 * the DSE.
	 */
	public void addConstraints(ExplorationSpecification specification) {
		constraints = specification.getTargets().stream()
				.filter(t -> (t instanceof ExplorationConstraint<?>))
				.map(ExplorationConstraint.class::cast).collect(Collectors.toList());
		constraints.forEach(c -> doAddConstraint(c));
	}

	private void doAddConstraint(ExplorationConstraint<?> constraint) {
		constraintViolationValue.put(constraint, 0.0);
		constraintViolationValueTracker.put(constraint, 0.0);
		constraintViolationTracker.put(constraint, 0);
	}

	public void addConstraintVioloation(ExplorationConstraint<?> violatedConstraint,
			double violationValue) throws Exception {

		// The ordering in calling these functions is important as to get the reset condition for
		// the problem dimensions correctly.
		// TODO: reset constraint violation if violationValue equals zero
		addConstraintViolationToConstraint(violatedConstraint, violationValue);
		addConstraintViolationToProblemDimension(violatedConstraint, violationValue);
	}

	public void updateConstraintViolationsForProblemDimensions() {
		// TODO: Original functionality is probably broken, rework during the transition to status
		// genotype.
		double violationValue = 0.0;
		int constraintViolationCount = 0;
		for(ExplorationConstraint<?> constraint : constraints) {
			Optional<IProblemDimension> probDim =
					Arrays.stream(constraint.getClass().getInterfaces())
							.filter(i -> IProblemDimension.class.isAssignableFrom(i.getClass()))
							.map(IProblemDimension.class::cast).findFirst();
			if(constraintViolationValue.containsKey(constraint)) {
				if(constraintViolationValue.get(constraint) > violationTolerance) {
					constraintViolationCount++;
				}
				violationValue += constraintViolationValue.get(constraint);
			}

			constraintViolationDegreePerProblemDimension.put(probDim.get(), violationValue);
			constraintViolationsPerProblemDimension.put(probDim.get(), constraintViolationCount);
		}
	}

	private void addConstraintViolationToConstraint(ExplorationConstraint<?> violatedConstraint,
			double violationValue) {
		int violationCount = 0;
		double constraintViolation = 0.0;
		double violationTrackerValue = 0.0;
		// TODO: Fix: Assume, the vales of the maps are always set jointly.
		if(constraintViolationValue.containsKey(violatedConstraint) &&
				constraintViolationValueTracker.containsKey(violatedConstraint) &&
				constraintViolationTracker.containsKey(violatedConstraint)) {
			constraintViolation = constraintViolationValue.get(violatedConstraint);
			violationTrackerValue = constraintViolationValueTracker.get(violatedConstraint);
			violationCount = constraintViolationTracker.get(violatedConstraint);
		}

		constraintViolation += violationValue;
		violationTrackerValue += violationValue;
		violationCount++;

		constraintViolationValue.put(violatedConstraint, constraintViolation);
		constraintViolationValueTracker.put(violatedConstraint, violationTrackerValue);
		constraintViolationTracker.put(violatedConstraint, violationCount);
	}

	private void addConstraintViolationToProblemDimension(
			ExplorationConstraint<?> violatedConstraint, double violationValue) throws Exception {
		EList<EClass> constraintSuperEClasses = violatedConstraint.eClass().getESuperTypes();
		EList<IProblemDimension> associatedProblemDimensionList =
				EcoreUtils.pickInstanceOf(IProblemDimension.class, constraintSuperEClasses);
		if(associatedProblemDimensionList.size() < 1) {
			// TODO: This WILL fail when the problem dimensions are implmented more dynamically, but
			// we should use the check for now.
			throw new Exception("Each Constraint type must have an IProlemDimension");
		}
		// Get(0) is safe due to the above check.
		IProblemDimension associatedProblemDimension = associatedProblemDimensionList.get(0);

		int constraintViolationCount = 0;
		double constraintViolationValue = 0.0;
		// TODO: Fix: Assume, constraintViolationsPerProblemDimension and
		// constraintViolationDegreePerProblemDimension are always set jointly.
		if(constraintViolationsPerProblemDimension.containsKey(associatedProblemDimension) &&
				constraintViolationDegreePerProblemDimension
						.containsKey(associatedProblemDimension)) {
			constraintViolationCount =
					constraintViolationsPerProblemDimension.get(associatedProblemDimension);
			constraintViolationValue =
					constraintViolationDegreePerProblemDimension.get(associatedProblemDimension);
		}

		constraintViolationCount++;
		constraintViolationValue += violationValue;

		constraintViolationsPerProblemDimension.put(associatedProblemDimension,
				constraintViolationCount);
		constraintViolationDegreePerProblemDimension.put(associatedProblemDimension,
				constraintViolationValue);
	}
}
