/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.create;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.ComposableDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.exception.DecodingException;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.ComposableGenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.graph.DecoderDependencyGraph;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.decode.ConstraintGenerationDecoder;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.opt4j.core.Genotype;
import org.opt4j.core.problem.Decoder;

/**
 * Base class for {@link ComposableCreator}s. It handles the decoding of the created
 * {@link Genotype} by launching the registered {@link Decoder}s.
 * 
 * @author diewald
 */
public abstract class ComposableCreatorBase<G extends ComposableGenotype>
		implements ComposableCreator<G> {
	/** References the {@link ComposableDecoder}s that decode the created {@link Genotype}. */
	private List<ComposableDecoder<G, ? extends Phenotype>> subDecoders = new ArrayList<>();

	/**
	 * Captures which {@link ComposableDecoder}s were already executed to avoid multiple executions
	 * of already executed decoders in the creation phase.
	 */
	private Map<ComposableDecoder<G, ? extends Phenotype>, Boolean> decoderExecuted =
			new HashMap<>();

	/**
	 * References the {@link DecoderDependencyGraph} which defines execution orders and which
	 * {@link Phenotype}s are needed within the DSE.
	 */
	private DecoderDependencyGraph depGraph;

	/** Default constructor that requires a representation of the input models. */
	public ComposableCreatorBase(DecoderDependencyGraph depGraph) {
		this.depGraph = depGraph;
	}

	/** {@inheritDoc} */
	@Override
	public void registerAssocDecoders(List<ComposableDecoder<G, ? extends Phenotype>> decoders) {
		subDecoders.addAll(decoders);
		decoders.forEach(d -> decoderExecuted.put(d, false));
	}

	/** {@inheritDoc} */
	@SafeVarargs
	@Override
	final public void registerAssocDecoders(ComposableDecoder<G, ?>... decoders) {
		for(ComposableDecoder<G, ?> dec : decoders) {
			@SuppressWarnings("unchecked") ComposableDecoder<G, ? extends Phenotype> castedDec =
					dec;
			subDecoders.add(castedDec);
			decoderExecuted.put(castedDec, false);
		}
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws DecodingException
	 */
	@SuppressWarnings("unchecked")
	@Override
	public <P extends Phenotype> FlatPhenotypeMap<Phenotype> decodeCreatedGenotype(G genotype,
			FlatPhenotypeMap<Phenotype> decodedPhenotypes,
			Set<Class<? extends Genotype>> createdGenotypeTypeSet) throws DecodingException {
		FlatPhenotypeMap<Phenotype> phenotypes = new FlatPhenotypeMap<>();
		Collection<Class<? extends IExplorationEncoding>> reqPhenotypeTypes =
				depGraph.getAllRequiredTypes();
		FlatPhenotypeMap<Phenotype> intermediatePhenos = new FlatPhenotypeMap<>();
		intermediatePhenos.addAll(decodedPhenotypes);

		Optional<? extends ComposableDecoder<G, P>> idDecoder =
				(Optional<? extends ComposableDecoder<G, P>>)subDecoders.stream().filter(
						d -> Arrays.asList(genotype.getClass()).containsAll(d.getRequiredTypes()))
						.findAny();

		for(ComposableDecoder<G, ? extends Phenotype> decoder : subDecoders) {
			/*
			 * if(!isHybridPhenoRequiredAndCreated(createdGenotypeTypeSet,
			 * decoder.getRequiredTypes()) ||
			 * Genotype.class.isAssignableFrom(decoder.getOutputType())) {
			 * continue;
			 * }
			 */

			ComposableDecoder<G, P> castedDecoder = (ComposableDecoder<G, P>)decoder;
			Class<P> decPhenotypeType = (Class<P>)decoder.getOutputType();
			// TODO: Here, the check with the Required Phenotypes should be removed, and replaced
			// with a check such that the required pheno/genotypes are available at this point.
			if(!decoderExecuted.get(decoder) && decoder != idDecoder.orElse(null) &&
					reqPhenotypeTypes.contains(decPhenotypeType) &&
					intermediatePhenos.containsAll(castedDecoder.getRequiredTypes())) {
				decoderExecuted.put(decoder, true);

				P decPhenotype = (P)decoder.decodeComposable(genotype, intermediatePhenos);
				// Null indicates a decoding error here.
				if(decPhenotype == null && !(decoder instanceof ConstraintGenerationDecoder)) {
					throw new DecodingException(genotype, decoder);
				}

				if(decPhenotype != null) {
					phenotypes.putEncoding(decPhenotype.getPTypeToRegister(), decPhenotype);
					intermediatePhenos.putEncoding(decPhenotype.getPTypeToRegister(), decPhenotype);
				}
			}
		}
		return phenotypes;
	}

	/** {@inheritDoc} */
	@Override
	public <P extends Phenotype> FlatPhenotypeMap<Phenotype> identityDecodeCreatedGenotype(
			G genotype, FlatPhenotypeMap<Phenotype> decodedPhenotypes) throws DecodingException {
		FlatPhenotypeMap<Phenotype> phenotypes = new FlatPhenotypeMap<>();
		FlatPhenotypeMap<Phenotype> intermediatePhenos = new FlatPhenotypeMap<>();
		intermediatePhenos.addAll(decodedPhenotypes);

		// If an identity decoder is present, execute it before any other decoder to ensure all
		// required data is available to the decoders.
		@SuppressWarnings("unchecked") Optional<? extends ComposableDecoder<G, P>> idDecoder =
				(Optional<? extends ComposableDecoder<G, P>>)subDecoders.stream().filter(
						d -> Arrays.asList(genotype.getClass()).containsAll(d.getRequiredTypes()))
						.findAny();
		if(idDecoder.isPresent()) {
			P decPhenotype = idDecoder.get().decodeComposable(genotype, intermediatePhenos);
			if(decPhenotype == null) {
				throw new DecodingException(genotype, idDecoder.get());
			}
			@SuppressWarnings("unchecked") Class<P> decType = (Class<P>)decPhenotype.getClass();
			phenotypes.putEncoding(decType, decPhenotype);
			intermediatePhenos.putEncoding(decType, decPhenotype);
		}

		return intermediatePhenos;
	}
}
