/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.selection;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.Opt4JDseTaskProvider;
import org.opt4j.core.optimizer.Operator;

import com.google.inject.Provider;

/**
 * "Key" class that enables registering both operator instances or their classes in the selection
 * map of the {@link IOperatorSelector}. Whenever an operator is bound by it's type, this class will
 * create an instance of it upon the first request.
 * 
 * @author diewald
 */
public class OperatorKey<O extends Operator<?>> {
	/**
	 * Provides the opt4j task that is used to construct an operator instance if it was registered
	 * by its type.
	 */
	private Provider<Opt4JDseTaskProvider> opt4JTaskProvider;

	/** The operator (cached in case of type registration). */
	private O operator;

	/** The operator type. Only used in case of type registration. */
	private Class<O> operatorType;

	/** Creates an {@link OperatorKey} for operator instances. */
	@SuppressWarnings("unchecked")
	public OperatorKey(O operator) {
		this.operator = operator;
		this.operatorType = (Class<O>)operator.getClass();
	}

	/**
	 * Creates an {@link OperatorKey} for operator type registrations. Requires the opt4j task
	 * provider for lazy {@link Operator} instantiation.
	 */
	public OperatorKey(Class<O> operatorType, Provider<Opt4JDseTaskProvider> provider) {
		this.opt4JTaskProvider = provider;
		this.operatorType = operatorType;
	}

	/** Returns the operator wrapped by this {@link OperatorKey}. */
	public O getOperator() {
		if(operator == null) {
			operator = opt4JTaskProvider.get().getInjector().getInstance(operatorType);
		}
		return operator;
	}

	// TODO: unsafe hash
	/** {@inheritDoc} */
	@Override
	public int hashCode() {
		// NOTE: We do not support multiple operators of the same type, but multiple operator types
		// that modify the same genotypes.
		return operatorType.hashCode();
	}

	/** {@inheritDoc} */
	@Override
	public boolean equals(Object other) {
		if(!(other instanceof OperatorKey)) {
			return false;
		}

		return this.hashCode() == other.hashCode();
	}
}
