/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.selection;

import java.util.HashMap;
import java.util.Map;

import org.opt4j.core.optimizer.Operator;
import org.opt4j.operators.AbstractGenericOperator;

import com.google.inject.AbstractModule;

/**
 * Base class for operator selection strategies. It is best suited for index-based strategies, such
 * as the ones provided by {@link RoundRobinOperatorSelector} and {@link RandomOperatorSelector}.
 * 
 * Since the set of all operators available for a particular operator type (e.g., mutation,
 * cross-over) is known to this class / the {@link IOperatorSelector} interface, also heuristics
 * based on the state of the optimization (e.g., current population) are possible (requires
 * acquisition of this information).
 * 
 * Classes providing a specific selection strategy will typically only override the
 * {@code *internal()} methods, whereas the methods inherited from {@link IOperatorSelector} serve
 * as interface to {@link OperatorSelectorPredicate}.
 * 
 * @author barner
 */
public abstract class OperatorSelectorBase<O extends Operator<?>> extends AbstractModule
		implements IOperatorSelector {

	/** Map: OperatorKey -> 0-based index */
	protected Map<OperatorKey<?>, Integer> operatorToIndexMap = new HashMap<>();

	/**
	 * Number of invocations of {@link #evaluate(OperatorKey)} in the current operator lookup loop
	 * in {@link AbstractGenericOperator#getOperator(org.opt4j.core.Genotype)}). Required to ensure
	 * that an operator is returned in any round (= |number of operators managed by this
	 * {@link IOperatorSelector}|, triggered by a single invocation of
	 * {@link AbstractGenericOperator#getOperator(org.opt4j.core.Genotype)}). In case an operator
	 * selection strategy would not select any operator in particular round, the last operator
	 * evaluated in that round will be returned.
	 */
	private int evaluationCount;

	/** {@inheritDoc} */
	@Override
	public void register(OperatorKey<? extends Operator<?>> operatorKey) {
		operatorToIndexMap.put(operatorKey, operatorToIndexMap.size());
		// registerInternal(operatorType);
	}

	/**
	 * Selection strategy specific registration of an operator. See
	 * {@link IOperatorSelector#register(OperatorKey)}.
	 * 
	 * @param operator
	 *            Operator to be registered.
	 */
	protected void registerInternal(Operator<? extends Operator<?>> operator) {
		// By default, do nothing
	}

	/**
	 * Selection strategy specific registration of an operator. See
	 * {@link IOperatorSelector#register(OperatorKey)}.
	 * 
	 * @param operator
	 *            Operator to be registered.
	 */
	protected void registerInternal(Class<Operator<?>> operator) {
		// By default, do nothing
	}

	/** {@inheritDoc} */
	@Override
	public final boolean evaluate(OperatorKey<? extends Operator<?>> operator) {

		int numOperators = operatorToIndexMap.size();

		// The actual evaluation if the current operator is to be selected
		boolean rval = evaluateInternal(operator.getOperator());

		evaluationCount++;
		if(rval || evaluationCount == numOperators) {
			// Operator has been selected, or this is the last operator to be investigated in this
			// round.

			// Reset
			evaluationCount = 0;
			finalizeSelectionRoundInternal();

			// Ensure that last operator is selected in any case
			rval = true;
		}

		return rval;
	}

	/** The actual selection strategy (invoked from {@link #evaluate(OperatorKey)}. */
	protected abstract boolean evaluateInternal(Operator<?> operator);

	/** Selection strategy finalization of a selection round. */
	protected abstract void finalizeSelectionRoundInternal();
}
