/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.comm;

import java.util.Collection;
import java.util.Map;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.comm.MessageEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.platform.PlatformCommunicationGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.AbstractTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.TaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FailSilentTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FailSilentTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.ComposableDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.Decodes;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.exception.DecodingException;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.comm.Message;
import org.fortiss.af3.exploration.alg.dse.sysmodel.comm.MessageGenerator;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.comm.IMessageRouter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.comm.IMessageRouter.RoutingStrategy;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.comm.MessageRouter;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;
import org.opt4j.core.Genotype;

import com.google.inject.Provides;

/**
 * {@link ComposableDecoder} that takes an {@link InstantiatedTaskMappingEncoding} and a) an
 * generates the required messages for this {@link InstantiatedTaskMappingEncoding} and b) routes
 * the generated {@link Message} on the target platform.
 * 
 * @author diewald
 */
public class MessageDecoder<S extends TaskMappingEntry<ITaskAdapter<?>, IExecutionUnitAdapter<?>>, T extends TaskMappingEncoding<ITaskAdapter<?>, S>>
		extends DecoderModule<MessageEncoding<S, T>> {
	/**
	 * {@link MessageGenerator} which analyzes an {@link InstantiatedTaskMappingEncoding} and
	 * determines the required {@link Message}s.
	 */
	protected MessageGenerator<S, T> messageGenerator;

	/** References the {@link IMessageRouter} selected in the Constructor. */
	protected IMessageRouter messageRouter;

	/** Default constructor that requires a reference to the input models. */
	public MessageDecoder() {
		messageGenerator = new MessageGenerator<S, T>();
		messageRouter = new MessageRouter();
	}

	/**
	 * Produces a {@link MessageEncoding} that contains the multicast messages that are required to
	 * exchange signals between {@link IExecutionUnitAdapter}s. For each signal that is exchanged
	 * between two tasks or more (only one sender) that are located on different
	 * {@link IExecutionUnitAdapter}s, a {@link Message} is generated.
	 * 
	 * @param tmEnc
	 *            {@link TaskMappingEncoding} defining the "location" of {@link ITaskAdapter}s.
	 * @param pcgEnc
	 *            Graph of the platform defining potential routes for {@link Message}s.
	 * @param atgEnc
	 *            {@link TaskGraphEncoding} containing the signals exchanged between
	 *            {@link ITaskAdapter}s.
	 * @return {@link MessageEncoding} that contains the multicast messages that are required to
	 *         exchange signals between {@link IExecutionUnitAdapter}s
	 * @throws DecodingException
	 */
	@SuppressWarnings("unchecked")
	@Provides
	@Decodes
	public MessageEncoding<FailSilentTaskMappingEntry, FailSilentTaskMappingEncoding> decode(
			FailSilentTaskMappingEncoding tmEnc, PlatformCommunicationGraphEncoding pcgEnc,
			AbstractTaskGraphEncoding atgEnc) throws DecodingException {
		return (MessageEncoding<FailSilentTaskMappingEntry, FailSilentTaskMappingEncoding>)decodeInternal(
				(T)tmEnc, pcgEnc, atgEnc);
	}

	/**
	 * Performs the actual decoding triggered by calling
	 * {@link #decode(FailSilentTaskMappingEncoding, PlatformCommunicationGraphEncoding, AbstractTaskGraphEncoding)}
	 * . This method is used since Guice doesn't allow bindings to undefined generics.
	 */
	public MessageEncoding<S, T> decodeInternal(T tmEnc, PlatformCommunicationGraphEncoding pcgEnc,
			AbstractTaskGraphEncoding atgEnc) throws DecodingException {
		Collection<Message> generatedMessages = messageGenerator.generateMessages(tmEnc, atgEnc);
		Map<Message, DefaultDirectedGraph<IResourceAdapter<?>, DefaultEdge>> messageRouteMap =
				messageRouter.calculateMessageRoutes(generatedMessages, pcgEnc,
						RoutingStrategy.SHORTEST_PATH);
		if(messageRouteMap == null) {
			// At least one *invalid* communication link is required by the given mapping =>
			// infeasible mapping.
			throw new DecodingException(tmEnc, this,
					"At least one message could not be routed from the sending unit to all its receivers.");
		}

		// Add the calculated routes to the generated Messages.
		for(Message msg : generatedMessages) {
			msg.setRoute(messageRouteMap.get(msg));
		}

		return new MessageEncoding<>(generatedMessages);
	}

	/** {@inheritDoc} */
	@Override
	public MessageEncoding<S, T> decode(Genotype genotype) {
		// TODO Auto-generated method stub
		return null;
	}
}
