/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.module.CreatorModule;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;
import org.fortiss.af3.exploration.alg.dse.module.EvaluatorModule;
import org.fortiss.af3.exploration.alg.service.IExplorationFeatureService;
import org.opt4j.core.Genotype;
import org.opt4j.core.problem.Creator;
import org.opt4j.core.problem.Decoder;
import org.opt4j.operators.OperatorModule;

import com.google.common.collect.ClassToInstanceMap;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import com.google.common.collect.MutableClassToInstanceMap;
import com.google.inject.AbstractModule;

/**
 * Implementation of the {@link IExplorationFeatureService}.
 * 
 * @author diewald
 */
public abstract class ExplorationFeature implements IExplorationFeature {

	/** Name of the {@link IExplorationFeature}. */
	private String name;

	/** Set referencing all register candidate solution types. */
	private Set<Class<? extends Phenotype>> phenotypes = new HashSet<>();

	/** Map associating Operators with the {@link Genotype}s they modify. */
	private Map<Class<? extends Genotype>, CreatorModule<?>> creatorGenoMap = new HashMap<>();

	/** Map associating Genot with the {@link Genotype}s they modify. */
	private Multimap<Class<? extends Genotype>, OperatorModule<?>> operatorGenoMap =
			HashMultimap.create();

	/**
	 * References all registered exploration modules. Differing between {@link Creator}s,
	 * {@link Decoder}s etc. is currently not needed.
	 */
	private ClassToInstanceMap<AbstractModule> registeredExpModules =
			MutableClassToInstanceMap.create();

	/** Constructor. */
	public ExplorationFeature(String name) {
		this.name = name;
	}

	/**
	 * Registers an optimization variable ({@code genotype}) with the DSE, along with its
	 * initializer ({@code creator}) and modificator ({@code operator}).
	 * 
	 * @param genotype
	 *            optimization variable type.
	 * @param creator
	 *            initializes the opt. variable.
	 * @param operators
	 *            modifications applied during the process iterations.
	 */
	public <G extends Genotype> void addVariable(Class<G> genotype, CreatorModule<G> creator,
			Collection<OperatorModule<?>> operators) {
		// Input check: Allow only fully specified variable registrations.
		if(genotype == null) {
			throw new RuntimeException(getClass() + ": The passed genotype is null!");
		} else if(creator == null) {
			throw new RuntimeException(getClass() + ": The passed creator is null!");
		} else if(operators == null || operators.isEmpty()) {
			throw new RuntimeException(getClass() + ": The passed operator set is empty or null!");
		}

		// Input check: Register modules only once.
		@SuppressWarnings("unchecked") Class<CreatorModule<G>> creatorClass =
				(Class<CreatorModule<G>>)creator.getClass();
		if(registeredExpModules.getInstance(creatorClass) != null) {
			throw new RuntimeException(getClass() + ": The module " + creatorClass.getSimpleName() +
					" has already been registered!");
		}
		creatorGenoMap.put(genotype, creator);
		registeredExpModules.putInstance(creatorClass, creator);

		for(OperatorModule<?> operator : operators) {
			@SuppressWarnings("unchecked") Class<OperatorModule<?>> operatorClass =
					(Class<OperatorModule<?>>)operator.getClass();
			if(registeredExpModules.getInstance(operatorClass) != null) {
				throw new RuntimeException(getClass() + ": The module " +
						operatorClass.getSimpleName() + " has already been registered!");
			}
			operatorGenoMap.put(genotype, operator);
			registeredExpModules.putInstance(operatorClass, operator);
		}
	}

	/**
	 * Registers a problem definition with the DSE process. The {@code phenotype} is equivalent to
	 * a candidate solution that is produced by the given {@code decoder}.
	 * 
	 * @param phenotype
	 *            candidate solution type.
	 * @param decoder
	 *            transforms variables (and/or other inputs) to candidate solutions.
	 */
	public <P extends Phenotype> void addProblem(Class<P> phenotype, DecoderModule<P> decoder) {
		// Input check: Allow only fully specified variable registrations.
		if(phenotype == null) {
			throw new RuntimeException(getClass() + ": The passed phenotype is null!");
		} else if(decoder == null) {
			throw new RuntimeException(getClass() + ": The passed decoder is null!");
		}

		// Input check: Register modules only once.
		@SuppressWarnings("unchecked") Class<DecoderModule<P>> decoderClass =
				(Class<DecoderModule<P>>)decoder.getClass();
		if(registeredExpModules.getInstance(decoderClass) != null) {
			throw new RuntimeException(getClass() + ": The module " + decoderClass.getSimpleName() +
					" has already been registered!");
		}

		phenotypes.add(phenotype);
		registeredExpModules.putInstance(decoderClass, decoder);
	}

	/**
	 * Registers the given {@code evaluator} to quantifies the quality of a solution by some metric.
	 * 
	 * @param evaluator
	 *            applies some metric to a candidate solution.
	 */
	public void addEvaluator(EvaluatorModule<?, ?, ?> evaluator) {
		// Input check: Allow only fully specified variable registrations.
		if(evaluator == null) {
			throw new RuntimeException(getClass() + ": The passed evaluator is null!");
		}

		// Input check: Register modules only once.
		@SuppressWarnings("unchecked") Class<EvaluatorModule<?, ?, ?>> evaluatorClass =
				(Class<EvaluatorModule<?, ?, ?>>)evaluator.getClass();
		if(registeredExpModules.getInstance(evaluatorClass) != null) {
			throw new RuntimeException(getClass() + ": The module " +
					evaluatorClass.getSimpleName() + " has already been registered!");
		}

		registeredExpModules.putInstance(evaluatorClass, evaluator);
	}

	/** {@inheritDoc} */
	@Override
	public Set<Class<? extends Genotype>> getGenotypes() {
		return creatorGenoMap.keySet();
	}

	/** {@inheritDoc} */
	@Override
	public Set<CreatorModule<?>> getCreators() {
		return registeredExpModules.values().stream().filter(m -> m instanceof CreatorModule)
				.map(CreatorModule.class::cast).collect(Collectors.toSet());
	}

	/** {@inheritDoc} */
	@Override
	public CreatorModule<?> getCreatorOf(Class<? extends Genotype> genotype) {
		return creatorGenoMap.get(genotype);
	}

	/** {@inheritDoc} */
	@Override
	public Set<OperatorModule<?>> getOperators() {
		return registeredExpModules.values().stream().filter(m -> m instanceof OperatorModule)
				.map(OperatorModule.class::cast).collect(Collectors.toSet());
	}

	/** {@inheritDoc} */
	@Override
	public Collection<OperatorModule<?>> getOperatorsOf(Class<? extends Genotype> genotype) {
		return operatorGenoMap.get(genotype);
	}

	/** {@inheritDoc} */
	@Override
	public Set<DecoderModule<?>> getDecoders() {
		return registeredExpModules.values().stream().filter(m -> m instanceof DecoderModule)
				.map(DecoderModule.class::cast).collect(Collectors.toSet());
	}

	/** {@inheritDoc} */
	@Override
	public Set<EvaluatorModule<?, ?, ?>> getEvaluators() {
		return registeredExpModules.values().stream().filter(m -> m instanceof EvaluatorModule)
				.map(EvaluatorModule.class::cast).collect(Collectors.toSet());
	}

	/** {@inheritDoc} */
	@Override
	public String getName() {
		return name;
	}
}
