/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.graph;

import static org.apache.commons.collections4.CollectionUtils.isEqualCollection;
import static org.fortiss.af3.exploration.util.ExplorationUtils.isDebugVerboseEnabled;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.stream.Collectors;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.ComposableDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.exception.MissingDependencyException;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.graph.display.ObjectJGraphTVisualizer;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.exploration.alg.util.GraphUtils;
import org.jgrapht.Graphs;
import org.jgrapht.graph.DirectedAcyclicGraph;
import org.opt4j.core.optimizer.IncompatibilityException;
import org.opt4j.core.problem.Creator;
import org.opt4j.core.problem.Decoder;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;

/**
 * Defines the Execution order of {@link Decoder}s / {@link Creator}s etc.
 * 
 * @author diewald
 */
// TODO: extend documentation.
// TODO: extend for the evaluator registration.
public abstract class DependencyGraph<V, T extends IDependencyModule<V, ? extends V>, E extends DependencyEdge<T>> {

	/** Internal graph representation for the execution order of s. */
	@SuppressWarnings("unchecked")
	protected DirectedAcyclicGraph<Class<? extends V>, E> dependencyDagGraph =
			new DirectedAcyclicGraph<Class<? extends V>, E>(
					(Class<? extends E>)DependencyEdge.class);

	/** Contains the calculated execution order for operators. */
	protected LinkedList<T> operatorExecOrder = new LinkedList<>();

	/** Type of the vertices of {@code this} dependency graph. */
	private Class<? extends V> vertexType;

	/** Collection of requested types. */
	private Set<Class<? extends V>> requestedTypes;

	/** Collection of all required types. */
	private Set<Class<? extends V>> requiredTypes = new HashSet<>();

	/** Operators which implement an identity operation. */
	private Multimap<Class<? extends V>, T> identityOperators = HashMultimap.create();

	/**
	 * Identity operators whose only purpose is to place an input encoding in the phenotype. Such a
	 * operator does not declare any other dependencies.
	 */
	private Map<Class<? extends V>, T> pureIdentityOperators = new HashMap<>();

	/** Set of {@link IExplorationEncoding} types that can be obtained from Creators. */
	protected Set<Class<? extends V>> inputEncodings = new HashSet<>();

	/** Constructor. */
	public DependencyGraph(Set<Class<? extends V>> reqTypeCollection,
			Class<? extends V> vertexType) {
		requestedTypes = reqTypeCollection;
		this.vertexType = vertexType;
	}

	/**
	 * Adds an initial {@link IExplorationEncoding} that is provided by {@link Creator}s to the
	 * dependency graph.
	 */
	public void addCreatedEncoding(Class<? extends V> inputType) {
		inputEncodings.add(inputType);
	}

	/**
	 * Adds the input types and the generated output types of the given
	 * operator to this {@link DependencyGraph}. They are connected by edges
	 * which represent the given operator.
	 * 
	 * @throws IllegalArgumentException
	 *             If adding the given operator to {@code this} graph would result
	 *             in circular dependencies in executing the operators.
	 */
	public void addOperator(T edgeOperator) throws IllegalArgumentException {
		Collection<Class<? extends V>> requiredTypes = getReqInputTypesFrom(edgeOperator);
		// TODO: add a proper exception.
		if(vertexType.isAssignableFrom(edgeOperator.getOutputType())) {
			@SuppressWarnings("unchecked") Class<V> outType =
					(Class<V>)edgeOperator.getOutputType();
			// Decoder graph. Add vertices, then edges.
			dependencyDagGraph.addVertex(outType);
			requiredTypes.forEach(p -> dependencyDagGraph.addVertex(p));
			if(requiredTypes.size() == 1 &&
					(getFirst(requiredTypes).orElse(null) == edgeOperator.getOutputType())) {
				pureIdentityOperators.put(getFirst(requiredTypes).get(), edgeOperator);
			}
			for(Class<? extends V> inputPheno : requiredTypes) {
				// Allow Identity Decoders.
				if(inputPheno.equals(edgeOperator.getOutputType())) {
					identityOperators.put(inputPheno, edgeOperator);
				} else {
					E edge = dependencyDagGraph.getEdge(outType, inputPheno);
					if(edge == null) {
						edge = dependencyDagGraph.addEdge(outType, inputPheno);
					}
					edge.addDecoder(edgeOperator);
				}
			}
		}
	}

	/**
	 * Validates whether all required inputs demanded by the given module are V types and
	 * returns a casted collection of the required inputs if the mentioned condition is met.
	 * Otherwise, an exception is thrown.
	 */
	@SuppressWarnings("unchecked")
	private Collection<Class<? extends V>> getReqInputTypesFrom(T module) {
		Collection<Class<? extends V>> castedInputTypes = new HashSet<>();
		for(Class<?> curInputType : module.getRequiredTypes()) {
			if(curInputType.isAssignableFrom(vertexType)) {
				throw new IncompatibilityException(
						"The required input type " + curInputType.getSimpleName() + " is not an " +
								vertexType.getSimpleName() + ".");
			}
			castedInputTypes.add((Class<? extends V>)curInputType);
		}
		return castedInputTypes;
	}

	/**
	 * Returns the calculated execution order of the operator such that the
	 * requested input types can be obtained.
	 * 
	 * @throws MissingDependencyException
	 *             if some required input type of a operator is
	 *             not provided by another one.
	 */
	public List<T> getOperatorExecOrder() throws MissingDependencyException {
		if(operatorExecOrder.isEmpty()) {
			calculateExecutionOrder();
		}
		return operatorExecOrder;
	}

	/**
	 * Calculates a feasible execution order of operators w.r.t. to the dependencies implicitly
	 * encoded in the input parameters.
	 * 
	 * @throws MissingDependencyException
	 *             if some required input type of a operator is
	 *             not provided by another one.
	 */
	private void calculateExecutionOrder() throws MissingDependencyException {
		// TODO: add comments to explain the resoning inside this method.
		@SuppressWarnings("unchecked") DirectedAcyclicGraph<Class<? extends V>, E> execOrderGraph =
				new DirectedAcyclicGraph<>((Class<? extends E>)DependencyEdge.class);
		Queue<Class<? extends V>> inputTypes = new LinkedList<>();
		Set<Class<? extends V>> traversedTypes = new HashSet<>();
		Set<E> requiredOperators = new HashSet<>();
		inputTypes.addAll(requestedTypes);
		while(!inputTypes.isEmpty()) {
			@SuppressWarnings("unchecked") Class<V> operatorInputType = (Class<V>)inputTypes.poll();
			requiredOperators.addAll(dependencyDagGraph.outgoingEdgesOf(operatorInputType));
			for(E edge : requiredOperators) {
				Class<? extends V> srcType = dependencyDagGraph.getEdgeTarget(edge);
				if(!traversedTypes.contains(srcType)) {
					inputTypes.add(srcType);
					traversedTypes.add(srcType);
				}
			}
		}

		// required Operators --> additional Identity decoders.
		Collection<Class<? extends V>> inTypes = requiredOperators.stream()
				.map(d -> dependencyDagGraph.getEdgeSource(d)).collect(Collectors.toList());
		getIdentityOperatorsOf(inTypes).entrySet()
				.forEach(e -> identityOperators.put(e.getKey(), e.getValue()));
		getIdentityOperatorsOf(requestedTypes).entrySet()
				.forEach(e -> identityOperators.put(e.getKey(), e.getValue()));

		for(E edge : requiredOperators) {
			Graphs.addEdgeWithVertices(execOrderGraph, dependencyDagGraph, edge);
		}

		// DEBUG code: only include if the dependency graph shall be debugged. Otherwise, two
		// windows for each individual and iteration will be produced.
		if(isDebugVerboseEnabled()) {
			ObjectJGraphTVisualizer<Class<? extends V>, E> graphViz =
					new ObjectJGraphTVisualizer<Class<? extends V>, E>(dependencyDagGraph);
			graphViz.displayGraph("Dependency graph");

		}

		if(isDebugVerboseEnabled()) {
			ObjectJGraphTVisualizer<Class<? extends V>, E> graphViz =
					new ObjectJGraphTVisualizer<Class<? extends V>, E>(execOrderGraph);
			graphViz.displayGraph("Dependency graph (Exec Order)");

		}

		calculateExecOrderList(execOrderGraph);

		checkMissingDependencies();
	}

	/**
	 * Each implementation of the {@link DependencyGraph} must provide a check for missing
	 * dependencies, sine there may be alternative ways to retrieve the required inputs, e.g., via
	 * the {@link SystemModelAdapter} in the {@link ComposableDecoder}s.
	 * 
	 * @throws MissingDependencyException
	 *             If any required input of some required module is missing.
	 */
	protected abstract void checkMissingDependencies() throws MissingDependencyException;

	/**
	 * Calculates the effective execution order of the operators and adds them to the internal list
	 * of {@code this} {@link DependencyGraph}.
	 */
	private void
			calculateExecOrderList(DirectedAcyclicGraph<Class<? extends V>, E> execOrderGraph) {
		Iterator<Class<? extends V>> encodingIter = execOrderGraph.iterator();
		while(encodingIter.hasNext()) {
			Class<? extends V> encoding = encodingIter.next();
			// Itentity Decoders must be launched as soon as possible to forward the produced
			// phenoytpes to the later steps.
			T identDec = pureIdentityOperators.get(encoding);
			if(identDec != null) {
				operatorExecOrder.addFirst(identDec);
			}
			Collection<E> launchableDecoders = execOrderGraph.outgoingEdgesOf(encoding);
			for(E decoderEdge : launchableDecoders) {
				// assert (decoderEdge.getElements().size() == 1);
				for(T dec : decoderEdge.getElements()) {
					if(!operatorExecOrder.contains(dec)) {
						operatorExecOrder.addFirst(dec);
					}
				}
			}
		}

		if(getSubExecListComparator() != null) {
			Iterator<T> operatorIt = operatorExecOrder.iterator();
			T nextOperator = operatorIt.next();
			while(operatorIt.hasNext()) {
				T curOperator = nextOperator;
				List<T> curSubList = new ArrayList<>();

				while(isEqualCollection(curOperator.getRequiredTypes(),
						nextOperator.getRequiredTypes())) {
					curSubList.add(nextOperator);
					if(operatorIt.hasNext()) {
						nextOperator = operatorIt.next();
					} else {
						break;
					}
				}

				Collections.sort(curSubList, getSubExecListComparator());
			}
		}
	}

	/**
	 * Returns a {@link Comparator} that is used to sort Operators which could be executed at the
	 * same time since their dependencies are satisfied. If no specialized {@link Comparator} is
	 * returned, these operations are executed in an arbitrary order.
	 * Here, the implementing classes must provide the {@link Comparator} as this ordering is
	 * implementation-specific in contrast to the pure resolution of dependencies.
	 */
	protected abstract Comparator<T> getSubExecListComparator();

	/**
	 * Returns the set of <i>all</i> vertices that must be processed to obtain
	 * the requested outputs.
	 */
	public Set<Class<? extends V>> getAllRequiredTypes() {
		if(requiredTypes.isEmpty()) {
			// FIXME: calculation is launched excessively and returns empty collections.
			calculateAllRequiredTypes();
		}
		Set<Class<? extends V>> reqInputSet = new HashSet<>();
		for(Class<? extends V> reqInType : requiredTypes) {
			reqInputSet.add(reqInType);
		}
		return reqInputSet;
	}

	/**
	 * Calculates the set of vertices that must be processed to obtain the
	 * requested outputs.
	 */
	private void calculateAllRequiredTypes() {
		for(Class<? extends V> reqType : requestedTypes) {
			requiredTypes.addAll(GraphUtils.getAllPredescesorsOf(dependencyDagGraph, reqType));
		}
		requiredTypes.addAll(requestedTypes);
	}

	/** Returns the set of all requested vertices. */
	public Set<Class<? extends V>> getAllRequestedTypes() {
		Set<Class<? extends V>> inputTypeSet = new HashSet<>();
		for(Class<? extends V> encType : requestedTypes) {
			if(Phenotype.class.isAssignableFrom(encType)) {
				inputTypeSet.add(encType);
			}
		}
		return inputTypeSet;
	}

	/**
	 * Returns a collection of identity operators which operate on the given input types. The set of
	 * operators to be checked equals the set of operators in the graph.
	 */
	private Map<Class<? extends V>, T>
			getIdentityOperatorsOf(Collection<Class<? extends V>> encodingTypes) {
		Map<Class<? extends V>, T> filteredIdentityOperators = new HashMap<>();
		for(E edge : dependencyDagGraph.edgeSet()) {
			Class<? extends V> srcType = dependencyDagGraph.getEdgeSource(edge);
			if((srcType == dependencyDagGraph.getEdgeTarget(edge)) &&
					encodingTypes.contains(srcType)) {
				assert (edge.getElements()
						.size() > 1) : "Only one identity decoder for each element type is supported.";
				filteredIdentityOperators.put(srcType, edge.getElements().stream().findAny().get());
			}
		}
		return filteredIdentityOperators;
	}

	// private Map<Class<? extends V>, T>
	// getPureIdentityOperators(Collection<Class<? extends V>> encodingTypes) {
	// for(Class<? extends V> encoding : encodingTypes) {
	//
	// }
	// }

	/**
	 * Comparator to allow ordering of operator lists by means of the execution
	 * order defined in a {@link DependencyGraph}.
	 */
	public static class DependencyComparator<V, T extends IDependencyModule<V, ? extends V>, E extends DependencyEdge<T>>
			implements Comparator<T> {

		/** List of the execution order of {@link ComposableDecoder}s. */
		List<T> execDecList;

		/**
		 * Constructor.
		 * 
		 * @throws MissingDependencyException
		 *             if some required input type of an operator
		 *             is not provided by another one.
		 */
		public DependencyComparator(DependencyGraph<V, T, E> execDepGraph)
				throws MissingDependencyException {
			execDecList = execDepGraph.getOperatorExecOrder();
		}

		/** {@inheritDoc} */
		@Override
		public int compare(T left, T right) {
			return Integer.compare(execDecList.indexOf(left), execDecList.indexOf(right));
		}
	}
}
