/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.model.time.impl;

import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.exploration.dsl.model.expressions.ArithmeticExpression;
import org.fortiss.af3.exploration.dsl.model.expressions.BooleanExpression;
import org.fortiss.af3.exploration.dsl.model.expressions.ConstraintExpression;
import org.fortiss.af3.exploration.dsl.model.operators.ArithmeticOp;
import org.fortiss.af3.exploration.dsl.model.patterns.ElementProperty;
import org.fortiss.af3.exploration.dsl.model.patterns.TwoPropertyValueComparison;
import org.fortiss.af3.exploration.dsl.model.types.NumberLiteral;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.time.DeadlineConstraint;

/**
 * Static method implementations for {@link DeadlineConstraint}s.
 * 
 * @author diewald
 */
public class DeadlineConstraintStaticImpl {

	/** Static implementation of the {@link DeadlineConstraint#getStartComponent()} method. */
	public static Component getStartComponent(DeadlineConstraint constr) {
		if(constr == null) {
			return null;
		}
		ElementProperty<Component, Double> startProp = getFirstElemProp(constr);
		if(startProp != null) {
			return startProp.getModelElement();
		}
		return null;
	}

	/**
	 * Static implementation of the {@link DeadlineConstraint#setStartComponent(Component)} method.
	 */
	public static void setStartComponent(DeadlineConstraint constr, Component component) {
		if(constr == null || component == null) {
			return;
		}
		ElementProperty<Component, Double> startProp = getFirstElemProp(constr);
		if(startProp != null) {
			startProp.setModelElement(component);
		}
	}

	/** Static implementation of the {@link DeadlineConstraint#getEndComponent()} method. */
	public static Component getEndComponent(DeadlineConstraint constr) {
		if(constr == null) {
			return null;
		}
		ElementProperty<Component, Double> startProp = getSecondElemProp(constr);
		if(startProp != null) {
			return startProp.getModelElement();
		}
		return null;
	}

	/**
	 * Static implementation of the {@link DeadlineConstraint#setEndComponent(Component)} method.
	 */
	public static void setEndComponent(DeadlineConstraint constr, Component component) {
		if(constr == null || component == null) {
			return;
		}
		ElementProperty<Component, Double> startProp = getSecondElemProp(constr);
		if(startProp != null) {
			startProp.setModelElement(component);
		}
	}

	/** Static implementation of the {@link DeadlineConstraint#getDeadline()} method. */
	public static Double getDeadline(DeadlineConstraint constr) {
		if(constr == null) {
			return null;
		}
		NumberLiteral<Double> deadline = getDeadlineNumberLiteral(constr);
		if(deadline != null) {
			return deadline.getValue();
		}
		return null;
	}

	/** Static implementation of the {@link DeadlineConstraint#setDeadline(Double)} method. */
	public static void setDeadline(DeadlineConstraint constr, Double period) {
		if(constr == null || period == null) {
			return;
		}
		NumberLiteral<Double> deadline = getDeadlineNumberLiteral(constr);
		if(deadline != null) {
			deadline.setValue(period);
		}
	}

	// FIXME: move the majority of the methods below to a util class for
	// TwoPropertyValueComparison(Pattern)!!!

	/**
	 * Returns the first {@link ElementProperty} (a {@link Component}) from a DeadlineConstraint.
	 */
	public static ElementProperty<Component, Double> getFirstElemProp(DeadlineConstraint constr) {
		if(constr == null) {
			return null;
		}
		ArithmeticOp<Double, ElementProperty<Component, Double>, ElementProperty<Component, Double>> arithOp =
				getLeftArithmeticOperator(constr);
		if(arithOp != null) {
			return arithOp.getLeft();
		}
		return null;
	}

	/**
	 * Returns the second {@link ElementProperty} (a {@link Component}) from a DeadlineConstraint .
	 */
	public static ElementProperty<Component, Double> getSecondElemProp(DeadlineConstraint constr) {
		if(constr == null) {
			return null;
		}
		ArithmeticOp<Double, ElementProperty<Component, Double>, ElementProperty<Component, Double>> arithOp =
				getLeftArithmeticOperator(constr);
		if(arithOp != null) {
			return arithOp.getRight();
		}
		return null;
	}

	/** Returns the left-side {@link ArithmeticOp} from a DeadlineConstraint . */
	@SuppressWarnings("unchecked")
	public static
			ArithmeticOp<Double, ElementProperty<Component, Double>, ElementProperty<Component, Double>>
			getLeftArithmeticOperator(DeadlineConstraint constr) {
		if(constr == null) {
			return null;
		}
		TwoPropertyValueComparison<Double, Component, Component> propValComp =
				getPropValCompExpression(constr);
		if(propValComp != null) {
			ArithmeticExpression<Double> aExpr = propValComp.getLeft();
			if(aExpr instanceof ArithmeticOp) {
				return (ArithmeticOp<Double, ElementProperty<Component, Double>, ElementProperty<Component, Double>>)aExpr;
			}
		}
		return null;
	}

	/**
	 * Returns the right hand side {@link NumberLiteral} (the deadline) from a DeadlineConstraint.
	 */
	public static NumberLiteral<Double> getDeadlineNumberLiteral(DeadlineConstraint constr) {
		if(constr == null) {
			return null;
		}
		TwoPropertyValueComparison<Double, Component, Component> propValComp =
				getPropValCompExpression(constr);
		if(propValComp != null) {
			ArithmeticExpression<Double> aExpr = propValComp.getRight();
			if(aExpr instanceof NumberLiteral) {
				return (NumberLiteral<Double>)aExpr;
			}
		}
		return null;
	}

	/**
	 * Returns the {@link TwoPropertyValueComparison} of a DeadlineConstraint from the containing
	 * {@link ExplorationConstraint}.
	 */
	public static TwoPropertyValueComparison<Double, Component, Component>
			getPropValCompExpression(DeadlineConstraint constr) {
		if(constr == null) {
			return null;
		}

		if(constr.getMoeaExpression() instanceof ConstraintExpression) {
			@SuppressWarnings("unchecked") ConstraintExpression<BooleanExpression> cExpr =
					(ConstraintExpression<BooleanExpression>)constr.getMoeaExpression();
			if(cExpr.getExpression() instanceof TwoPropertyValueComparison) {
				// Type Parameter enforced by access methods.
				@SuppressWarnings("unchecked") TwoPropertyValueComparison<Double, Component, Component> compExpr =
						(TwoPropertyValueComparison<Double, Component, Component>)cExpr
								.getExpression();
				return compExpr;
			}
		}
		return null;
	}
}
